type TemplateMap = Map<string, { template: any, mapper: (...args: any) => any }>;

class SolsticeWidget<Options extends { templateId: string }> {
  element: HTMLElement;
  options: Options;
  templates: TemplateMap;
  fetches: Array<(options: Options) => any>;

  constructor(element: HTMLElement, options: Options, templates: TemplateMap, fetches: Array<(options: Options) => any>) {
    this.element = element;
    this.options = options;
    this.templates = templates;
    this.fetches = fetches;
  }

  async fetch() {
    const data = this.fetches.map(f => f(this.options));
    return Promise.all(data);
  }

  async render() {
    // Templates can be given at class instantiation.
    const data = await this.fetch();
    console.log(data);
    this.templates.get(this.options.templateId).template(data);
  }
}

export class SolsticeWidgetBuilder<Options extends { templateId: string }> {
  selector: string;
  templates: Map<string, { template: any, mapper: (...args: any) => any }>;
  fetches: Array<(options: Options) => any>;
  optionParser: (dataset: HTMLElement['dataset']) => Options;

  constructor(selector: string) {
    this.selector = selector;
    this.templates = new Map();
    this.fetches = new Array();
  }

  setOptionParser(callbackFn: (dataset: HTMLElement['dataset']) => Options) {
    this.optionParser = callbackFn;
    return this;
  }

  createTemplate(name: string, template: string, mapper: (...args: any) => any) {
    this.templates.set(name, { template, mapper });
    return this;
  }

  fetch(callbackFn: (options: Options) => any) {
    this.fetches.push(callbackFn);
    return this;
  }

  build() {
    const elements = Array.from(document.querySelectorAll(this.selector)) as HTMLElement[];
    elements.map((element) => { 
      const options = this.optionParser(element.dataset);
      return new SolsticeWidget(element, options, this.templates, this.fetches) 
    });
  }
}
